""" 
# Available Modules
Available modules are listed in the `fennol.models.modules.MODULES` dictionary.
They can also be found by searching for 'FID' in the documentation's search bar.
"""

from .fennix import FENNIX
from .modules import register_fennix_module, get_modules_documentation, available_fennix_modules,MODULES
